/**
 * Copyright 2016 the WHATAP project authors. All rights reserved.
 * Use of this source code is governed by a license that
 * can be found in the LICENSE file.
 */

var MeterSocketio = require('../counter/meter/meter-socket.io'),
    TraceContextManager = require('../trace/trace-context-manager'),
    SocketStep = require('../step/socket-step'),
    conf = require('../conf/configure'),
    IPUtil = require('../util/iputil'),
    Logger = require('../logger');
const {callback} = require("pg/lib/native/query");
const {Detector: URLPatternDetector} = require("whatap/lib/trace/serviceurl-pattern-detector");
const HashUtil = require("whatap/lib/util/hashutil");
const DataTextAgent = require("whatap/lib/data/datatext-agent");
const ResourceProfile = require("whatap/lib/util/resourceprofile");
const MeterUsers = require("whatap/lib/counter/meter/meter-users");
const UserIdUtil = require("whatap/lib/util/userid-util");
const MessageStep = require("whatap/lib/step/message-step");
const Hexa32 = require("whatap/lib/util/hexa32");
const ProfilePack         = require('../pack/profile-pack');
const TxRecord            = require('../service/tx-record');
const DateUtil            = require('../util/dateutil');
const SecurityMaster      = require('../net/security-master');
const DataProfileAgent    = require('../data/dataprofile-agent');
const MeterService        = require('../counter/meter/meter-service').MeterService;

var SocketIOObserver = function(agent){
    this.agent = agent;
    this.packages = ['socket.io'];
};

SocketIOObserver.prototype.inject = function (mod, moduleName) {
    if (mod.__whatap_observe__) {
        return;
    }
    mod.__whatap_observe__ = true;
    Logger.initPrint("SocketIOObserver");
    if (conf.trace_background_socket_enabled === false) {
        return;
    }

    var self = this;
    var aop = self.agent.aop;

    aop.before(mod.Server?.prototype, 'on', function (obj, args, ret, lctx) {
        args[args.length-1] = self.__createTransactionObserver(args[args.length-1]);
    })
};

SocketIOObserver.prototype.__createTransactionObserver = function (callback) {
    var self = this;

    return function (){
        TraceContextManager._asyncLocalStorage.run(initCtx(), () => {
            var ctx = TraceContextManager._asyncLocalStorage.getStore();
            if(!ctx) {
                return;
            }

            MeterSocketio.connected(1);

            var host;
            const socket = arguments[0];
            const address = socket.handshake.address;
            if(address && address.includes(':')){
                host = address.substring(address.lastIndexOf(':')+1);
            }

            ctx.socket_connecting = true;

            ctx.footprint('Socket Connecting: ' + host);

            var step = new SocketStep();
            step.start_time = ctx.getElapsedTime();
            step.ipaddr = Buffer.from(IPUtil.stringToBytes(host));
            // step.port = port;

            ctx.socket_connecting = false;
            step.elapsed = ctx.getElapsedTime() - step.start_time;
            ctx.profile.push(step);

            ctx.footprint('Socket Connecting Done');

            self.__endTransaction(null, ctx);

            return;
        })
    }
}

SocketIOObserver.prototype.__endTransaction = function(error, ctx) {
    try {
        var profile = new ProfilePack();
        var wtx = new TxRecord();
        wtx.endTime = DateUtil.currentTime();
        profile.time  = wtx.endTime;
        wtx.elapsed = ctx.getElapsedTime();

        ctx.service_hash = HashUtil.hashFromString(ctx.service_name);
        DataTextAgent.SERVICE.add(ctx.service_hash, ctx.service_name);

        wtx.service = ctx.service_hash;
        wtx.cpuTime = ResourceProfile.getCPUTime() - ctx.start_cpu;
        wtx.malloc = ResourceProfile.getUsedHeapSize()-ctx.start_malloc;
        if(wtx.malloc < 0) { wtx.malloc = 0; }
        wtx.status = 2;

        MeterService.add(wtx.service, wtx.elapsed,
            wtx.errorLevel, ctx.mcaller_pcode, ctx.mcaller_okind, ctx.mcaller_oid);

        profile.oid = SecurityMaster.OID;
        profile.service = wtx;

        TraceContextManager.end(ctx._id);
        setTimeout(function () {
            DataProfileAgent.sendProfile(ctx, profile, false);
            ctx = null;
        }, 100);

    } catch (e) {
        Logger.printError('WHATAP-607', 'End transaction error..', e, false);
        TraceContextManager.end(ctx._id);
        ctx = null;
    }

};

function initCtx() {
    const ctx = TraceContextManager.start();
    if (!ctx) {
        return;
    }

    ctx.start_malloc = ResourceProfile.getUsedHeapSize();
    ctx.start_cpu = ResourceProfile.getCPUTime();

    var remote_addr = IPUtil.checkIp4("0.0.0.0");
    ctx.remoteIp = IPUtil.stringToInt(remote_addr);

    return ctx;
}

exports.SocketIOObserver = SocketIOObserver;